Set Implicit Arguments.
Require Import MyTactics.
Require Import Heaps.
Require Import HeapResources.

(* ---------------------------------------------------------------------------- *)

(* Agreement between a memory block and a heap access right. *)

(* The value [v] contained in the field is required to agree with the access
   right [r]. The adopter pointer is ignored; see [AdoptionAgreement]. *)

Definition block_agrees (b : block) (r : heap_access_right) :=
  let (_, v) := b in
  dx_agrees eq eq v r.

Hint Unfold block_agrees : agrees.

Ltac block_agrees_inversion :=
  unfold block_agrees in *; dx_agrees_inversion.

(* ---------------------------------------------------------------------------- *)

(* Agreement between heaps and instrumented heaps. *)

Definition heap_agrees (h : heap) (R : heap_resource) :=
  aMap_agrees block_agrees h R.

Hint Unfold heap_agrees : agrees.

(* ---------------------------------------------------------------------------- *)

(* If [R] describes some machine state, then [R] is consistent. *)

Lemma block_agrees_consistent:
  forall b r,
  block_agrees b r ->
  consistent r.
Proof.
  unfold block_agrees. destruct b. eauto with consistent.
Qed.

Hint Resolve block_agrees_consistent : consistent.

Goal forall h R, heap_agrees h R -> consistent R.
Proof.
  eauto with consistent.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The empty heap agrees with the empty instrumented heap. *)

Lemma heap_agrees_void:
  heap_agrees empty_heap heap_void.
Proof.
  unfold empty_heap. eauto with agrees.
Qed.

Hint Resolve heap_agrees_void : agrees.

(* ---------------------------------------------------------------------------- *)

(* Agreement between the physical heap and the block resource is not
   affected by updates to the adopter pointers. *)

Lemma heap_agrees_pointer_update:
  forall h R l p1 p2 v,
  heap_agrees h R ->
  maps h l (Block p1 v) ->
  heap_agrees (update h l (Block p2 v)) R.
Proof.
  intros.
  forwards: @agrees_maps_left_to_right; [ eauto | eauto | unpack ].
  equates 1. eauto with agrees. symmetry. eauto using @update_identity.
Qed.

Hint Resolve heap_agrees_pointer_update : agrees.

