Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Syntax.
Require Import MSADXPermissions.

(* ---------------------------------------------------------------------------- *)

(* A subset of types and permissions are duplicable. *)

(* One must not confuse the syntactic judgement that is defined here, which
   concerns permissions and types, with the semantic notion of [duplicable]
   that is defined in [MSA], which concerns resources. *)

CoInductive dup : term -> Prop :=
| DupForall:
    forall kappa P,
    dup P ->
    dup (TyForall kappa P)
| DupExists:
    forall kappa P,
    dup P ->
    dup (TyExists kappa P)
| DupSingleton:
    forall v,
    dup (TySingleton v)
| DupArrow:
    forall T U,
    dup (TyArrow T U)
| DupMixStar:
    forall T P,
    dup T ->
    dup P ->
    dup (TyMixStar T P)
| DupAnchored:
    forall v T,
    dup T ->
    dup (TyAnchored v T)
| DupEmpty:
    dup TyEmpty
| DupStar:
    forall P Q,
    dup P ->
    dup Q ->
    dup (TyStar P Q)
| DupDup:
    (* The permission [TyDuplicable T] is duplicable. Interesting kind
       of bootstrap. *)
    forall T,
    dup (TyDuplicable T)
| DupNil:
    dup RowNil
| DupCons:
    forall T Ts,
    dup T ->
    dup Ts ->
    dup (RowCons T Ts)
| DupLock:
    forall P,
    dup (TyLock P)
| DupRef:
    forall T,
    dup T ->
    dup (TyRef Duplicable T)
| DupAdoptable:
    dup TyAdoptable
.

Hint Constructors dup : dup.

Ltac select_dup action :=
  match goal with h: dup _ |- _ => action h end.
Ltac dup_inversion :=
  gji select_dup.

(* ---------------------------------------------------------------------------- *)

(* The judgement [dup] is stable by lifting. *)

Lemma dup_weakening:
  forall T,
  dup T ->
  forall x,
  dup (shift x T).
Proof.
  cofix. inversion 1; intros; simpl_lift_goal; eauto with dup.
Qed.

Hint Resolve dup_weakening : j_weakening.

(* ---------------------------------------------------------------------------- *)

(* The judgement [dup] is stable by substitution. *)

Lemma dup_substitution:
  forall T,
  dup T ->
  forall w x,
  dup (subst w x T).
Proof.
  cofix. inversion 1; intros; simpl_subst_goal; eauto with dup.
Qed.

Hint Resolve dup_substitution : j_substitution.

(* ---------------------------------------------------------------------------- *)

(* The type [TyTop] is duplicable. *)

Lemma dup_TyTop:
  dup TyTop.
Proof.
  eauto with dup.
Qed.

