Set Implicit Arguments.
Require Export MSA.
Require Export AllocationMaps.
Require Export MSAAllocationMaps.
Require Export MSADXPermissions.
Require Import Syntax.
Require Import Heaps.

(* ---------------------------------------------------------------------------- *)

(* A heap resource is a partial and instrumented heap. It maps each memory
   location to a (nothing-or-duplicable-or-exclusive) access right, which
   contains information about what data is stored at this location and
   indicates in what ways we are allowed to affect this data. Note that
   this access right concerns the regular fields only; the adopter field
   is special and is described by an adoption resource, which is another
   component of a resource. *)

Notation heap_access_right :=
  (dx term term).

Notation heap_resource :=
  (aMap heap_access_right).

Notation heap_void :=
  (empty_map _).

(* ---------------------------------------------------------------------------- *)

(* Heap resources form a monotonic separation algebra. (Sanity check.) *)

Goal MSA heap_resource.
Proof.
  eauto with typeclass_instances.
Qed.

(* ---------------------------------------------------------------------------- *)
