Set Implicit Arguments.
Require Import MyTactics.
Require Import Syntax.
Require Import WellKindedness.

(* The fact that all kinds are inhabited seems to be required in order to
   obtain the strengthening lemma. Indeed, we have existential types. If there
   was an uninhabited kind [kappa], then, in a non-empty kind environment, one
   could build a value of type [TyExists kappa (shift 0 T)], yet one would not
   be able to transport this type derivation down to an empty kind
   environment, where there is no longer a witness of kind [kappa]. *)

(* We can restrict this lemma to the substitutable kinds. Indeed, it is
   used in conjunction with the substitution lemma, which is restricted
   to substitutable kinds already. *)

Lemma all_substitutable_kinds_inhabited:
  forall kappa,
  substitutable kappa ->
  exists T,
  wk zilch T kappa.
Proof.
  introv h. destruct kappa;
  try solve [ dependent destruction h; unfold logic in *; tauto ].
  (* VALUE *)
  exists TUnit. eauto with wk.
  (* PERM *)
  exists TyEmpty. eauto with wk.
  (* TYPE *)
  exists TyTop. eauto with wk.
  (* ROW _ *)
  exists RowNil. eauto with wk.
Qed.

