Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Typing.
Require Import JTop.

(* Classification. *)

(* When we have a canonical type derivation about a value, which exhibits a
   known type constructor, what can we say of the value? *)

(* ------------------------------------------------------------------------- *)

(* Functions. *)

(* If [v] has a function type, then [v] must be a function. *)

Lemma classification_TyArrow:
  forall R K v T U,
  jctop R K v (TyArrow T U) ->
  exists t,
  v = TAbs t.
Proof.
  jctop_simplified_induction; eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has a memory block type, then [v] must be a memory location. *)

Lemma classification_TyRef:
  forall R K v m T,
  jctop R K v (TyRef m T) ->
  exists l,
  v = TLoc l.
Proof.
  jctop_simplified_induction; eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has a lock type, then [v] must be a lock address. *)

Lemma classification_TyLock:
  forall R K v P,
  jctop R K v (TyLock P) ->
  exists lck,
  v = TLock lck.
Proof.
  jctop_simplified_induction; eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has an adoptee type, then [v] must be a memory location. *)

Lemma classification_TyAdoptable: 
  forall R K v, 
  jctop R K v TyAdoptable -> 
  exists l, 
  v = TLoc l. 
Proof. 
  jctop_simplified_induction; eauto. 
Qed. 

Lemma classification_TyUnadopted:
  forall R K v,
  jctop R K v TyUnadopted ->
  exists l,
  v = TLoc l.
Proof.
  jctop_simplified_induction; eauto.
Qed.

(* ------------------------------------------------------------------------- *)

(* If [v] has an adopter type, then [v] must be a memory location. *)

Lemma classification_TyAdopts:
  forall R K v U,
  jctop R K v (TyAdopts U) ->
  exists l,
  v = TLoc l.
Proof.
  jctop_simplified_induction; eauto.
Qed.


(* ------------------------------------------------------------------------- *)

(* The following tactic applies one of the above classification lemmas. *)

Ltac classification :=
  repeat match goal with
  | h: jctop _ _ _ (TyArrow _ _) |- _ =>
      generalize (classification_TyArrow h); intro;
      unpack; subst; wk_inversion;
      revert h
  | h: jctop _ _ _ (TyRef _ _) |- _ =>
      generalize (classification_TyRef h); intro;
      unpack; subst;
      revert h
  | h: jctop _ _ _ (TyLock _) |- _ =>
      generalize (classification_TyLock h); intro;
      unpack; subst;
      revert h
  | h: jctop _ _ _ TyAdoptable |- _ => 
      generalize (classification_TyAdoptable h); intro; 
      unpack; subst; 
      revert h 
  | h: jctop _ _ _ TyUnadopted |- _ =>
      generalize (classification_TyUnadopted h); intro;
      unpack; subst;
      revert h
  | h: jctop _ _ _ (TyAdopts _) |- _ =>
      generalize (classification_TyAdopts h); intro;
      unpack; subst;
      revert h
  end;
  intros.

