Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Typing.
Require Import JBasic.
Require Import JAffinity.
Require Import JStability.
Require Import Agreement.

(* ---------------------------------------------------------------------------- *)

(* The judgement [js R t] states that the thread soup [t] is well-typed under
   the resources [R]. *)

Inductive js : resource -> term -> Prop :=

(* A single thread [TThread t] is well-typed under [R] if the term [t] is. *)

  | JsThread:
      forall n R t T,
      jt n R zilch TyEmpty t T ->
      js R (TThread t)

(* The construct [TPar t u] requires the soups [t] and [u] to be well-typed
   under independent (and compatible) resources. *)

  | JsPar:
      forall R R1 R2 t1 t2,
      star R1 R2 = R ->
      js R1 t1 ->
      js R2 t2 ->
      js R (TPar t1 t2).

Hint Constructors js : js.

(* ---------------------------------------------------------------------------- *)

(* The judgement [jconf c] means that the configuration [c] is well-typed. *)

(* [jconf] is the global invariant that holds throughout execution. It requires
   that the state [s] and the term [t] be well-formed and be well-typed under a
   common resource [R]. *)

Inductive jconf : configuration -> Prop :=
| JConf:
    forall R s t,
    wkc (Cf s t) SOUP ->
    jsr s R ->
    js R t ->
    jconf (Cf s t).

(* ---------------------------------------------------------------------------- *)

(* If the source program [t] is well-typed, then the initial configuration,
   which consists of a single thread, about to execute [t], in the initial
   state, is well-typed. *)

Lemma jconf_initial:
  forall n t T,
  jt n void zilch TyEmpty t T ->
  jconf (Cf initial (TThread t)).
Proof.
  econstructor.
    eauto using wks_initial with wk.
    eauto using jsr_initial.
    eauto with js.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The judgement [js] is affine. *)

Lemma js_drop_right:
  forall R1 R2 t,
  js R1 t ->
  consistent (star R1 R2) ->
  js (star R1 R2) t.
Proof.
  induction 1; intros; subst.
  (* JsThread *)
  eauto using j_drop_right with js.
  (* JsPar *)
  (* Use the induction hypothesis on just one side (whichever). *)
  econstructor; [ | eapply IHjs1; eauto with consistent_aac | eauto ].
  aac_reflexivity.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The judgement [js] is stable. *)

Lemma js_rely:
  forall R1 t,
  js R1 t ->
  consistent R1 ->
  forall R2,
  rely R1 R2 ->
  js R2 t.
Proof.
  induction 1; intros; subst;
  try rely_star_clear;
  try econstructor (solve [
    eauto using (get1 j_rely) with consistent_split
  ]).
Qed.

