Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Dup.
Require Import Harmless.
Require Import Heaps.
Require Import Resources.
Require Import Typing.
Require Import JBasic.

(* ---------------------------------------------------------------------------- *)

(* The predicate [invariant _ lck P] is stable. *)

Lemma invariant_stable:
  forall R1 R2 lck P,
  invariant R1 lck P ->
  rely R1 R2 ->
  invariant R2 lck P.
Proof.
  intros. eapply mapsp_stablep; eauto with stable.
Qed.

(* The predicate [taken _ lck] is stable. *)

Lemma taken_stable:
  forall R1 R2 lck,
  taken R1 lck ->
  rely R1 R2 ->
  taken R2 lck.
Proof.
  intros. eapply mapsp_stablep; eauto with stable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The predicate [adoptable] is stable. *)

Lemma adoptable_stable:
  forall R1 R2 l,
  adoptable R1 l ->
  rely R1 R2 ->
  adoptable R2 l.
Proof.
  intros. eapply mapsp_stablep; eauto with stable. 
Qed.

(* The predicate [unadopted _ l] is stable. *)

Lemma unadopted_stable:
  forall R1 R2 l,
  unadopted R1 l ->
  rely R1 R2 ->
  unadopted R2 l.
Proof.
  intros. eapply mapsp_stablep; eauto with stable.
Qed.

(* The predicate [all_adoptees (get _) xs y] is stable. *)

Lemma all_adoptees_get_stable:
  forall (R1 R2 : resource) xs y,
  all_adoptees (get R1) xs y ->
  rely R1 R2 ->
  all_adoptees (get R2) xs y.
Proof.
  intros. explode_resources_adoption_strong. eauto using all_adoptees_stable.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The typing judgements are monotonic with respect to [rely]: actions by other
   threads cannot invalidate the permissions held by the current thread. *)

Lemma j_rely:
(
  forall f n R1 K P t T,
  j f n R1 K P t T ->
  consistent R1 ->
  forall R2,
  rely R1 R2 ->
  j f n R2 K P t T
) /\ (
  forall f n R1 K P tag Tss ts T,
  jbs f n R1 K P tag Tss ts T ->
  consistent R1 ->
  forall R2,
  rely R1 R2 ->
  jbs f n R2 K P tag Tss ts T
) /\ (
  forall R1 K P,
  jpi R1 K P ->
  consistent R1 ->
  forall R2,
  rely R1 R2 ->
  jpi R2 K P
).
Proof.
  apply j_jbs_jpi_mutind; intros; subst;
  (* Clear the original type derivation, which can mislead [eauto]
     when attempting to construct a new type derivation. *)
  repeat match goal with
  | h: j _ _ _ _ _ _ _ |- _ => clear h
  | h: jpi _ _ _ |- _ => clear h
  end;
  try rely_star_clear;
  eauto 6 using @consistent_core, @rely_core,
                @maps_stable, invariant_stable, taken_stable,
                adoptable_stable, unadopted_stable, all_adoptees_get_stable, adopted_stable
           with consistent_split stable explode_resources j.
Qed.

