Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import Inhabited.

(* ---------------------------------------------------------------------------- *)

(* The typing judgements are stable by strengthening, which is the inverse of
   weakening. Whereas weakening adds an unused variable to the environment,
   strengthening removes such a variable. *)

(* We specialize the lemma to [jpi], because that is all we need. *)

Lemma jpi_strenthening:
  forall R K x kappa P,
  jpi R (insert x kappa K) (shift x P) ->
  substitutable kappa ->
  jpi R K P.
Proof.
  (* The strengthening lemma is a direct consequence of the substitution
     lemma. We could attempt to prove it by induction, but that would not
     allow us to get rid of the inhabitation hypothesis, I believe, so
     it is preferable to go this way. *)
  intros.
  forwards: all_substitutable_kinds_inhabited; [ eauto | unpack ].
  eapply j_substitution; eauto using wk_empty with subst_lift.
Qed.

(* ---------------------------------------------------------------------------- *)

(* We would now like to construct an iterated version of the above lemma (and
   specialize it for the case where [P] is closed). However, we must keep track
   of the fact that [K] contains substitutable kinds only. We could define a
   predicate that says ``every kind in [K] is substitutable''. However, it is
   just as simple, and more direct, to say ``[K] supports strengthening''. *)

Definition supports_strengthening K :=
  forall R P,
  jpi R K P ->
  closed 0 P ->
  jpi R zilch P.

Lemma supports_strengthening_zilch:
  supports_strengthening zilch.
Proof.
  unfold supports_strengthening. eauto.
Qed.

Lemma supports_strengthening_insert:
  forall K kappa,
  supports_strengthening K ->
  substitutable kappa ->
  supports_strengthening (insert 0 kappa K).
Proof.
  unfold supports_strengthening. introv h. intros.
  eapply h; eauto.
  eapply jpi_strenthening; eauto.
  erewrite @closed_lift_invariant; eauto with typeclass_instances.
Qed.

Hint Resolve supports_strengthening_zilch supports_strengthening_insert :
supports_strengthening.

