Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Zap.
Require Import Sugar.
Require Import WellKindedness.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.

(* ---------------------------------------------------------------------------- *)

(* Some typing rules have premises that involve closed values. These premises
   are stable under weakening, but this requires a special argument. *)

Lemma jctop_closed_substitution: (* useful for [JRef] *)
  forall R K v T U x,
  jpi R K (TyAnchored (subst U x v) T) ->
  wk zilch v VALUE ->
  jpi R K (TyAnchored v T).
Proof.
  intros. closed. assumption.
Qed.

Hint Resolve jctop_closed_substitution : j_substitution.

(* ---------------------------------------------------------------------------- *)

(* The typing judgements are stable by substitution. *)

(* This statement is restricted to substitutable kinds because it relies on
   [harmless_subst], which needs this restriction. *)

Lemma j_substitution:
(
  forall n K P Q,
  jps n K P Q ->
  forall w x kappa K' P' Q',
  insert x kappa K' = K ->
  subst w x P = P' ->
  subst w x Q = Q' ->
  wk K' w kappa ->
  substitutable kappa ->
  jps n K' P' Q'
) /\ (
  forall n K T1 T2,
  sub n K T1 T2 ->
  forall w x kappa K',
  insert x kappa K' = K ->
  wk K' w kappa ->
  substitutable kappa ->
  sub n K' (subst w x T1) (subst w x T2)
) /\ (
  forall n K Ts1 Ts2,
  row_sub n K Ts1 Ts2 ->
  forall w x kappa K',
  insert x kappa K' = K ->
  wk K' w kappa ->
  substitutable kappa ->
  row_sub n K' (subst w x Ts1) (subst w x Ts2)
) /\ (
  forall f n R K P t T,
  j f n R K P t T ->
  forall w x kappa K' P' t' T',
  insert x kappa K' = K ->
  subst w x P = P' ->
  subst w x t = t' ->
  subst w x T = T' ->
  wk K' w kappa ->
  substitutable kappa ->
  j f n R K' P' t' T'
) /\ (
  forall f n R K P tag Tss ts T,
  jbs f n R K P tag Tss ts T ->
  wkbs Tss ->
  forall w x kappa K' P' ts' T',
  insert x kappa K' = K ->
  (
    forall tag Ts,
    wk zilch Ts (ROW TYPE) ->
    subst w x (P tag Ts) = P' tag Ts
  ) ->
  subst w x ts = ts' ->
  subst w x T = T' ->    
  wk K' w kappa ->
  substitutable kappa ->
  jbs f n R K' P' tag Tss ts' T'
) /\ (
  forall R K P,
  jpi R K P ->
  forall w x kappa K' P',
  insert x kappa K' = K ->
  subst w x P = P' ->
  wk K' w kappa ->
  substitutable kappa ->
  jpi R K' P'
).
Local Hint Extern 1
  (subst _ _ (TyAnchored _ _)  = _)  => (simpl_subst_goal; f_equal).
Local Hint Extern 1
  (subst _ _ (TyStar _ _)      = _)  => (simpl_subst_goal; f_equal).
Local Hint Extern 1
  (subst _ _ (TyArrow _ _)     = _)  => (simpl_subst_goal; f_equal).
Local Hint Extern 1
  (subst _ _ (TyAdopts _)     = _)  => (simpl_subst_goal; f_equal).
Local Hint Extern 1
  (TyAnchored _ _ = TyAnchored _ _)  => (f_equal).
Local Hint Extern 1
  (List.map _ _ = _) => simpl; simpl_subst_goal.
Proof.
  apply the_great_mutind; intros; subst; simpl_subst_goal; try closed;
  try econstructor (solve [
    eauto 7 with insert_insert insert_concat lift_subst subst_subst j_substitution
  ]).
  (* JForallIntro *)
  (* The lemma [harmless_subst] requires a well-kindedness fact which is
     slightly non-obvious, so we treat this case by hand. *)
  econstructor.
    eapply harmless_subst; eauto using wk_weakening_inverse with wk.
    eauto 7 with insert_insert insert_concat lift_subst subst_subst j_substitution.
Qed.

(* The following tactic applies the substitution lemma (in forward mode)
   to a typing judgement [j] that appears among the hypotheses. *)

Ltac substitution :=
  let ht := fresh "ht" in
  forwards ht: (get4 j_substitution); eauto with substitutable;
  instantiate; repeat rewrite subst_lift in ht.

