Set Implicit Arguments.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Zap.
Require Import Sugar.
Require Import WellKindedness.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Typing.
Require Import JBasic.

(* ---------------------------------------------------------------------------- *)

(* Some typing rules have premises that involve closed values. These premises
   are stable under weakening, but this requires a special argument. In some
   cases, this argument can be carried out on the fly by the tactic [closed];
   in other cases, we must prepare a lemma for use by [eauto]. *)

Lemma jctop_closed_weakening: (* useful for [JRef] *)
  forall R K v T x,
  jpi R K (TyAnchored (shift x v) T) ->
  wk zilch v VALUE ->
  jpi R K (TyAnchored v T).
Proof.
  intros. closed. assumption.
Qed.

Hint Resolve jctop_closed_weakening : j_weakening.

(* ---------------------------------------------------------------------------- *)

(* The typing judgements are stable by lifting. *)

Lemma j_weakening:
(
  forall n K P1 P2,
  jps n K P1 P2 ->
  forall x okappa K' P1' P2',
  raw_insert x okappa K = K' ->
  shift x P1 = P1' ->
  shift x P2 = P2' ->
  jps n K' P1' P2'
) /\ (
  forall n K T1 T2,
  sub n K T1 T2 ->
  forall x okappa K',
  raw_insert x okappa K = K' ->
  sub n K' (shift x T1) (shift x T2)
) /\ (
  forall n K Ts1 Ts2,
  row_sub n K Ts1 Ts2 ->
  forall x okappa K',
  raw_insert x okappa K = K' ->
  row_sub n K' (shift x Ts1) (shift x Ts2)
) /\ (
  forall f n R K P t T,
  j f n R K P t T ->
  forall x okappa K' P' t' T',
  raw_insert x okappa K = K' ->
  shift x P = P' ->
  shift x t = t' ->
  shift x T = T' ->
  j f n R K' P' t' T'
) /\ (
  forall f n R K P tag Tss ts T,
  jbs f n R K P tag Tss ts T ->
  wkbs Tss ->
  forall x okappa K' P' ts' T',
  raw_insert x okappa K = K' ->
  (
    forall tag Ts,
    wk zilch Ts (ROW TYPE) ->
    shift x (P tag Ts) = P' tag Ts
  ) ->
  shift x ts = ts' ->
  shift x T = T' ->    
  jbs f n R K' P' tag Tss ts' T'
) /\ (
  forall R K P,
  jpi R K P ->
  forall x okappa K' P',
  raw_insert x okappa K = K' ->
  shift x P = P' ->
  jpi R K' P'
).
Local Hint Extern 1
  (shift _ (TyAnchored _ _)  = _)   => (simpl_lift_goal; f_equal).
Local Hint Extern 1
  (shift _ (TyStar _ _)      = _)   => (simpl_lift_goal; f_equal).
Local Hint Extern 1
  (shift _ (TyArrow _ _)     = _)   => (simpl_lift_goal; f_equal).
Local Hint Extern 1
  (TyAnchored _ _ = TyAnchored _ _) => f_equal.
Local Hint Extern 1
  (subst _ 0 _ = subst _ 0 _) => f_equal.
Local Hint Extern 1
  (List.map _ _ = _) => simpl; simpl_lift_goal.
Proof.
  apply the_great_mutind; intros; subst; simpl_lift_goal; try closed;
  try econstructor (solve [
    eauto with lookup_insert insert_insert insert_concat lift_lift lift_subst j_weakening
  ]).
Qed.

(* ---------------------------------------------------------------------------- *)

(* If [R] satisfies a closed permission [P] under the empty environment, then
   [R] satisfies [P] under an arbitrary environment [K]. *)

Lemma jpi_zilch_weakening:
  forall R P,
  jpi R zilch P ->
  forall K,
  jpi R K P.
Proof.
  introv h. assert (wk zilch P PERM). eauto with wk.
  (* This is an iterated version of the above weakening lemma, specialized for
     the judgement [jpi], and specialized for a closed permission [P]. *)
  induction K.
  assumption.
  eapply j_weakening.
    eauto.
    eapply raw_insert_zero.
    closed. reflexivity.
Qed.

