Set Implicit Arguments.
Require Import MyTactics.
Require Import LockHeaps.
Require Import LockResources.

(* Agreement between lock heaps and lock resources. *)

(* ---------------------------------------------------------------------------- *)

(* We first define agreement at the level of a single lock. We require the bit
   found in the concrete lock heap and in the bit found in the lock resource
   to coincide. (One might think that [LockZero] means ``either unlocked or
   locked by some other thread'', and, as a result, that one must allow
   [LOCKED] to agree with [LockZero]. However, we are only talking about
   complete resources here -- the resource represents the combination of the
   views of all threads -- so it is possible to require exact agreement.) *)

Inductive lock_agrees : lock_status -> lock_info -> Prop :=
| LockAgreesUnlocked:
    forall P,
    lock_agrees UNLOCKED (LockZero P)
| LockAgreesLocked:
    forall P,
    lock_agrees LOCKED (LockOne P).

Hint Constructors lock_agrees : agrees.

Ltac lock_agrees_inversion :=
  match goal with h: lock_agrees _ _ |- _ => inversion h; clear h; subst end.

(* ---------------------------------------------------------------------------- *)

(* This notion is then lifted up to the level of allocation maps. *)

Definition lock_heap_agrees (h : lock_heap) (R : lock_resource) :=
  aMap_agrees lock_agrees h R.

Hint Unfold lock_heap_agrees : agrees.

(* ---------------------------------------------------------------------------- *)

(* If [R] describes some machine state, then [R] is consistent. *)

Lemma lock_agrees_consistent:
  forall s R,
  lock_agrees s R ->
  consistent R.
Proof.
  inversion 1; subst; eauto with consistent explode.
Qed.

Hint Resolve lock_agrees_consistent : consistent.

Goal forall h R, lock_heap_agrees h R -> consistent R.
Proof.
  eauto with consistent.
Qed.

(* ---------------------------------------------------------------------------- *)

(* The empty lock heap agrees with the empty lock resource. *)

Lemma lock_heap_agrees_void:
  lock_heap_agrees empty_lock_heap lock_void.
Proof.
  unfold empty_lock_heap. eauto with agrees.
Qed.

Hint Resolve lock_heap_agrees_void : agrees.

(* ---------------------------------------------------------------------------- *)

