Set Implicit Arguments.
Generalizable All Variables.
Require Import MyTactics.
Require Import MSA.
Require Import MSASig.

(* ------------------------------------------------------------------------- *)

(* Example. *)

(* Let's pick the monotonic separation algebra of the natural integers as
   the base algebra. *)

Require Import MSANat.
Transparent msa_nat. (* Just for us. *)

(* Let's focus on the property of being even. *)

Definition even (n : nat) :=
  exists k, n = 2 * k.

(* This property is compatible with the base algebra. *)

Program Instance msasig_even: MSASig even.

Next Obligation. (* The sum of two even integers is even. *)
  unfold even in *. unpack. subst. simpl.
  match goal with x: nat, y: nat |- _ => exists (x + y) end.
  omega.
Qed.

Next Obligation. (* The unit, zero, is even. *)
  simpl. exists 0. omega.
Qed.

Next Obligation. (* Equality preserves evenness. *)
  simpl in *. subst. assumption.
Qed.

(* So, the system can prove that [sig even], the type of even integers,
   forms a monotonic separation algebra. *)

Goal MSA (sig even).
Proof.
  eauto with typeclass_instances.
Qed.

