Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import Agreement.
Require Import JConf.
Require Import ProgressTerms.

(* ---------------------------------------------------------------------------- *)

(* A thread soup is acceptable if and only if every thread either: (i) has
   finished (with an answer) or (ii) is waiting on some lock or (iii) is able
   to make progress in the current state. In other words, no thread has
   crashed. *)

Inductive acceptable : configuration -> Prop :=
  | AcceptableThreadAnswer:
      forall s t,
      answer t ->
      acceptable (Cf s (TThread t))
  | AcceptableThreadWaiting:
      forall s t,
      waiting s t ->
      acceptable (Cf s (TThread t))
  | AcceptableThreadRed:
      forall s1 t1 s2 t2,
      (* Note: a single thread reduces, but it may reduce to a configuration
         of more than one thread. *)
      red (Cf s1 (TThread t1)) (Cf s2 t2) ->
      acceptable (Cf s1 (TThread t1))
  | AcceptablePar:
      forall s t1 t2,
      acceptable (Cf s t1) ->
      acceptable (Cf s t2) ->
      acceptable (Cf s (TPar t1 t2)).

Hint Constructors acceptable : acceptable.

(* ---------------------------------------------------------------------------- *)

(* The main progress lemma for configurations of kind [SOUP]. *)

Lemma progress_SOUP:
  forall Rt t,
  js Rt t ->
  forall Rf R s,
  wkc (Cf s t) SOUP ->
  jsr s R ->
  star Rt Rf = R ->
  acceptable (Cf s t).
Proof.
  induction 1; inversion 1; intros; subst; wk_inversion.

  (* TThread *)
  (* By the progress lemma for terms, four cases arise. *)
  forwards [ ? | [ hs | [ ? | ? ]]]: progress_TERM; eauto.
  (* Sub-case: [t] is an answer. *)
  solve [ eauto with acceptable ].
  (* Sub-case: [t] spawns a new thread. Then, [TThread t] reduces. *)
  destruct hs; subst.
  solve [ eauto with acceptable red ].
  (* Sub-case: [t] is waiting. *)
  solve [ eauto with acceptable ].
  (* Sub-case: [t] takes a step. Then, [TThread t] takes a step too. *)
  unpack. destruct_configurations. solve [ eauto with acceptable red ].

  (* TPar *)
  solve [ eauto 6 with acceptable wk aac_guess ].

Qed.

(* ---------------------------------------------------------------------------- *)

(* Progress for configurations. A simplified statement. *)

Theorem progress:
  forall c,
  jconf c ->
  acceptable c.
Proof.
  inversion 1; intros; subst. eauto using progress_SOUP, @core_right_unit.
Qed.

