Set Implicit Arguments.
Require Import MyList.
        Import ListNotations.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JStrengthening.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JStability.
Require Import JDuplication.
Require Import JTop.
Require Import JDecomposition.
Require Import JClassification.
Require Import SubsumptionSoundness.
Require Import JCanonicalization.
Require Import HiddenInvariants.
Require Import Agreement.
Require Import SubjectReductionAuxiliary.

(* ---------------------------------------------------------------------------- *)

(* Subject reduction for [TNewLock]. *)

Lemma subject_reduction_TNewLock:
  forall s1 s2 t2,
  redwkc (Cf s1 TNewLock) (Cf s2 t2) TERM ->
  forall R R',
  jsr s1 (star R R') ->
  forall K P,
  (K <> zilch -> harmless TNewLock) ->
  wk K P PERM ->
  sr_target R' K (Cf s2 t2) (intersection [
    TyLock P;
    TyLocked
  ]).
Proof.
  intros. redwkc_inversion. jsr_consistent.
  destruct_states. explode_resources_locks.
  eapply SrTargetCanonical.

  (* Sub-goal 1. The new lock heap agrees with [R2]. *)
  { eapply jsr_allocate_lock; eauto with reflexivity. }

  (* Sub-goal 2. [R2] and [R2f] together form the new global resource. *)
  { simpl. explode; eauto using @star_allocate_core_right_unit. }

  (* Sub-goal 3. The framed-out resource [R2f] evolves along [rely]. *)
  { explode; eauto using @rely_allocate_core, @rely_reflexive with consistent. }

  (* Sub-goal 4. The new lock is well-typed. *)
  eapply JcTopIntersection. eauto with wk. simpl.
  (* Split the lock resource between the two permissions that we must
     justify (lemma [invariant_taken_split]). We also split the heap
     resource, but how it is split is irrelevant here, as it is not
     consulted. *)
  eapply JpiStar.
  3: explode; first [ eapply invariant_taken_split | eapply core_left_unit ].

  (* Sub-sub-goal 4.1. The permission [TyLock P]. *)
  jctop JLock.
    (* This is the key place where we need to argue that [P] is closed.
       A priori, all we know is that [P] has kind [PERM] under [K].
       However, because [TNewLock] is not harmless, we know that we
       cannot be under a Lambda-abstraction, hence [K] must be empty. *)
    { replace zilch with K. assumption. eauto using harmful_TNewLock. }
    { eapply new_lock_invariant_core; eauto with reflexivity consistent_split limit_limit. }

  (* Sub-sub-goal 4.2. The permission [TyLocked]. *)
  eapply JpiStarEmpty.
  jctop JLocked.
    eapply new_lock_taken; eauto with reflexivity consistent_split limit_limit.

Qed.

