Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JStrengthening.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JStability.
Require Import JDuplication.
Require Import JTop.
Require Import JDecomposition.
Require Import JClassification.
Require Import SubsumptionSoundness.
Require Import JCanonicalization.
Require Import HiddenInvariants.
Require Import Agreement.
Require Import SubjectReductionAuxiliary.

(* ---------------------------------------------------------------------------- *)

(* Subject reduction for [TRead]. *)

Lemma subject_reduction_TRead:
  forall s1 v s2 t2,
  redwkc (Cf s1 (TRead v)) (Cf s2 t2) TERM ->
  forall R R',
  jsr s1 (star R R') ->
  forall K m T,
  jpi R K (TyStar (TyDuplicable T) (TyAnchored v (TyRef m T))) ->
  sr_target R' K (Cf s2 t2) (TyMixStar T (TyAnchored v (TyRef m T))).
Proof.
  intros. redwkc_inversion. jsr_consistent. decomposition.
  (* The memory location [l] is allocated in the heap and in the current
     resource. Furthermore, the heap and the current resource agree.
     Draw the consequences. *)
  jsr_maps_maps.
  (* The value [v] found at address [l] is well-typed under some resource [Rv].
     Because this value admits a duplicable type, this derivation is in fact
     valid under [core Rv]. *)
  forwards hcore: j_duplicable; [ eauto | eauto with consistent_aac | eauto with dup | ].
  (* In this case, there is no side effect. The heap remains well-typed. *)
  (* Using [jctop_jt] in this way allows us to construct a [jctop]
     judgement for the reduct, instead of a [jt] judgement. *)
  forwards: jctop_jt; [ | unpack; sr_target_pure ].
  (* Drop a resource that we do not need. *)
  jpi_drop_left.
  (* Apply the frame rule. *)
  eapply JcTopFrame.
    (* In the first premise, consume [core Rv]. *)
    eexact hcore.
    (* In the second premise, consume [Rv]. *)
    clear hcore. jctop JRef; eauto.
    (* Because [core Rv] and [Rv] combine to produce [Rv], this
       application of the frame rule is valid. *)
    match goal with Rv: resource |- _ => aac_rewrite (core_left_unit Rv) end.
    reflexivity.
Qed.

