Set Implicit Arguments.
Require Import MyLogic.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Environments.
Require Import Syntax.
Require Import Sugar.
Require Import WellKindedness.
Require Import Heaps.
Require Import LockHeaps.
Require Import Configurations.
Require Import EvaluationContexts.
Require Import Reduction.
Require Import Dup.
Require Import Harmless.
Require Import Resources.
Require Import Row.
Require Import Typing.
Require Import JBasic.
Require Import JWeakening.
Require Import JSubstitution.
Require Import JStrengthening.
Require Import DerivedSubtyping.
Require Import DerivedTyping.
Require Import JAffinity.
Require Import JStability.
Require Import JDuplication.
Require Import JTop.
Require Import JDecomposition.
Require Import JClassification.
Require Import SubsumptionSoundness.
Require Import JCanonicalization.
Require Import HiddenInvariants.
Require Import Agreement.
Require Import SubjectReductionAuxiliary.

(* ---------------------------------------------------------------------------- *)

(* Subject reduction for [TRelease]. *)

Lemma subject_reduction_TRelease:
  forall s1 v s2 t2,
  redwkc (Cf s1 (TRelease v)) (Cf s2 t2) TERM ->
  forall R R',
  jsr s1 (star R R') ->
  forall K P,
  supports_strengthening K ->
  jpi R K (TyStar P (TyStar (TyAnchored v TyLocked) (TyAnchored v (TyLock P)))) ->
  sr_target R' K (Cf s2 t2) TyTop.
Proof.
  intros. redwkc_inversion. jsr_consistent. decomposition.
  destruct_states. explode_resources_locks.
  (* This is what we have. *)
  forwards: invariant_taken; [ eauto | eauto | eauto with consistent_split | ].
  (* When releasing the lock, we must give up its invariant. *)
  forwards: jsr_release; [
    eauto
  | eauto
  | (* defer until the meta-variables are resolved *)
  | (* The strengthening lemma is required here. We know that [R] justifies
       [P] under the empty environment; hence, this holds also under [K]. *)
    unfold supports_strengthening in *;
    eauto using wk_closed
  | explode; aac_guess
  | simpl; eauto
  | simpl; eauto
  |
  ].
  (* The subgoal that was deferred. *)
  simpl. maps_maximal.
  (* Split. *)
  sr_target_pure.
  (* Sub-goal 2. *)
  explode; try aac_guess.
    erewrite @star_update_solitary_left.
    f_equal. aac_guess.
    eauto with maps_maximal.
    eauto with solitary.
    eauto with consistent_aac.
    eauto.
  (* Sub-goal 4. *)
  eauto using JTop with wk.
Qed.

