Set Implicit Arguments.
Require Import List.
Require Import MyTactics.
Require Import DeBruijn.
Require Import Syntax.

(* If [xs] is a list of memory locations and [U] is a type, then [zap xs U] is
   the iterated conjunction [(TLoc xs) @ U]. *)

Fixpoint zap (xs : list nat) (U : term) : term :=
  match xs with
  | nil =>
      TyEmpty
  | x :: xs =>
      TyStar (TyAnchored (TLoc x) U) (zap xs U)
  end.

Lemma zap_weakening:
  forall xs x U,
  shift x (zap xs U) = zap xs (shift x U).
Proof.
  induction xs; simpl; intros; simpl_lift_goal; eauto with f_equal.
Qed.

Hint Resolve zap_weakening : j_weakening.

Lemma zap_substitution:
  forall xs w x U,
  subst w x (zap xs U) = zap xs (subst w x U).
Proof.
  induction xs; simpl; intros; simpl_subst_goal; eauto with f_equal.
Qed.

Hint Resolve zap_substitution : j_substitution.

