Set Implicit Arguments.
Require Import ProofIrrelevance.

(* This is a variant of the lemmas [subset_eq_compat] and [subsetT_eq_compat]
   defined in the standard library module [ProofIrrelevanceFacts]. *)

Lemma subset_eq_compat:
  forall (U : Type) (P : U -> Prop) (x y : U) (p : P x) (q : P y),
  x = y ->
  exist P x p = exist P y q.
Proof.
  (* The proof is simpler than the one used in the standard library.
     Thanks to Xavier Leroy for suggesting the use of [subst]. *)
  intros. subst. f_equal. eapply proof_irrelevance.
Qed.

(* This tactic simplifies all hypotheses of the form [exist _ _ _ = exist _ _ _]. *)

Ltac exist_exist_hypotheses :=
  repeat match goal with h: exist _ ?x _ = exist _ ?y _ |- _ =>
    assert (x = y); [ congruence | clear h ]
  end.

(* This tactic simplifies a goal of the form [exist _ _ _ = exist _ _ _]. *)

Ltac exist_exist_goal :=
  try eapply subset_eq_compat.

(* The following lemma does not exploit proof irrelevance, and could be useful
   when dealing with a goal of the form [exist P a1 p1 = exist P a2 ?p2], where
   [?p2] is a meta-variable. Applying the lemma should have the effect of
   instantiating [?p2]. Unfortunately, the lemma is tricky to use, as [apply]
   will not produce [a1 = a2] as a subgoal, but as another meta-variable. *)

Lemma exist_exist_left_to_right:
  forall A (P : A -> Prop) a1 a2 p,
  forall h : a1 = a2,
  exist P a1 p = exist P a2 (eq_rect _ _ p _ h).
Proof.
  intros. subst. simpl. reflexivity.
Qed.

Lemma exist_exist_right_to_left:
  forall A (P : A -> Prop) a1 a2 p,
  forall h : a1 = a2,
  exist P a1 (eq_rect _ _ p _ (eq_sym h)) = exist P a2 p.
Proof.
  intros. subst. simpl. reflexivity.
Qed.

